@@ -1,10 +1,11 @@
 from __future__ import annotations
-from typing import Any, Callable, Container, Deque, Dict, Iterable, Iterator, List, NoReturn, Optional, Reversible, Sequence, Set, Sized, TYPE_CHECKING, Tuple, Type, Generic, Union
+from typing import Any, Callable, Container, Deque, Dict, Iterable, Iterator, List, NoReturn, Optional, Reversible, Sequence, Set, Sized, TYPE_CHECKING, Type, Generic, Union
 
 if TYPE_CHECKING:
     from .lookup import Lookup
     from .grouping import Grouping
     from .ordered_enumerable import OrderedEnumerable
+    from .cached_enumerable import CachedEnumerable
 
 from .types_linq_error import InvalidOperationError, IndexOutOfRangeError
 from .more_typing import (
@@ -12,7 +13,6 @@ from .more_typing import (
     TDefault,
     TInner,
     TKey,
-    TOther,
     TResult,
     TSource_co,
     TValue,
@@ -30,18 +30,21 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
     def __init__(self,
         it: Union[Iterable[TSource_co], Callable[[], Iterable[TSource_co]]]
     ):
-        if callable(it):
-            self._iter_factory = lambda it_=it: it_()
-        else:
+        if isinstance(it, Iterable):
             self._iter_factory = lambda it_=it: it_
+        else:
+            self._iter_factory = lambda it_=it: it_()
+
+    def _get_iterable(self) -> Iterable[TSource_co]:
+        return self._iter_factory()
 
     # 'fallback = F' -> calls dunder methods if available
     #                  -> otherwise calls own implementation
     # 'fallback = T'-> calls own implementation
     def _contains_impl(self, value: object, fallback: bool) -> bool:
-        iterable = self._iter_factory()
+        iterable = self._get_iterable()
         if not fallback and isinstance(iterable, Container):
-            return value in iterable
+            return value in iterable  # type: ignore
         for elem in iterable:
             if elem == value:
                 return True
@@ -57,7 +60,7 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
         index: Union[int, slice],
         fallback: bool,
     ) -> Union[TSource_co, Enumerable[TSource_co]]:
-        iterable = self._iter_factory()
+        iterable = self._get_iterable()
         if isinstance(index, int):
             # Sequence is an abstract base class without @runtime_checkable
             if not fallback and isinstance(iterable, Sequence):
@@ -109,16 +112,16 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
                         .reverse()._every(-step)
             return Enumerable(inner)
 
-    def __getitem__(self,
+    def __getitem__(self,  # type: ignore[override]
         index: Union[int, slice],
     ) -> Union[TSource_co, Enumerable[TSource_co]]:
         return self._getitem_impl(index, fallback=False)
 
     def __iter__(self) -> Iterator[TSource_co]:
-        return iter(self._iter_factory())
+        return iter(self._get_iterable())
 
     def _len_impl(self, fallback: bool) -> int:
-        iterable = self._iter_factory()
+        iterable = self._get_iterable()
         if not fallback and isinstance(iterable, Sized):
             return len(iterable)
         count = 0
@@ -129,7 +132,7 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
         return self._len_impl(fallback=False)
 
     def _reversed_impl(self, fallback: bool) -> Iterator[TSource_co]:
-        iterable = self._iter_factory()
+        iterable = self._get_iterable()
         # Sequence is an abstract base class without @runtime_checkable
         if not fallback and isinstance(iterable, (Sequence, Reversible)):
             return reversed(iterable)
@@ -192,6 +195,10 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
             yield element
         return Enumerable(inner)
 
+    def as_cached(self, *, cache_capacity: Optional[int] = None) -> CachedEnumerable[TSource_co]:
+        from .cached_enumerable import CachedEnumerable
+        return CachedEnumerable(self, cache_capacity)
+
     def _average_helper(self, selector, when_empty):
         count = 0
         iterator = iter(self)
@@ -222,7 +229,7 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
         return self._average_helper(selector, lambda: default)
 
     def cast(self, _: Type[TResult]) -> Enumerable[TResult]:
-        return self
+        return self  # type: ignore
 
     def concat(self, second: Iterable[TSource_co]) -> Enumerable[TSource_co]:
         def inner():
@@ -264,7 +271,7 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
                 yield default
                 return
             yield from iterator
-        return Enumerable(inner)
+        return Enumerable(inner)  # type: ignore
 
     def distinct(self) -> Enumerable[TSource_co]:
         def inner():
@@ -496,7 +503,7 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
         else:  # len(args) == 2:
             comparer = None
         return OrderedEnumerable(
-            self._iter_factory,
+            self._get_iterable,
             None,
             key_selector,
             comparer,
@@ -513,7 +520,7 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
         else:  # len(args) == 2:
             comparer = None
         return OrderedEnumerable(
-            self._iter_factory,
+            self._get_iterable,
             None,
             key_selector,
             comparer,
@@ -536,13 +543,13 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
         if count is not None:
             if count < 0:
                 Enumerable._raise_count_negative()
-            def inner(curr=start, cnt=count):  # type: ignore
+            def inner(curr=start, cnt=count):  # type: ignore[misc]
                 while cnt > 0:
                     yield curr
                     curr += 1
                     cnt -= 1
         else:
-            def inner(curr=start):
+            def inner(curr=start):  # type: ignore[misc]
                 while True:
                     yield curr
                     curr += 1
@@ -553,12 +560,12 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
         if count is not None:
             if count < 0:
                 Enumerable._raise_count_negative()
-            def inner(val=value, cnt=count):  # type: ignore
+            def inner(val=value, cnt=count):  # type: ignore[misc]
                 while cnt > 0:
                     yield val
                     cnt -= 1
         else:
-            def inner(val=value):
+            def inner(val=value):  # type: ignore[misc]
                 while True:
                     yield val
         return Enumerable(inner)
@@ -802,7 +809,7 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
             value_selector: Any = lambda x: x
         else:  # len(args) == 1
             value_selector = args[0]
-        return dict((key_selector(e), value_selector(e)) for e in self)
+        return {key_selector(e): value_selector(e) for e in self}
 
     def to_set(self) -> Set[TSource_co]:
         return {e for e in self}
@@ -847,17 +854,17 @@ class Enumerable(Sequence[TSource_co], Generic[TSource_co]):
                     yield elem
         return Enumerable(inner)
 
-    def zip(self,
-        second: Iterable[TOther],
-        *args: Callable[[TSource_co, TOther], TResult],
-    ) -> Union[Enumerable[TResult], Enumerable[Tuple[TSource_co, TOther]]]:
-        if len(args) == 0:
-            result_selector: Any = lambda x, y: (x, y)
-        else:  # len(args) == 1
-            result_selector = args[0]
+    def zip(self, *iters: Iterable[Any]) -> Enumerable[Any]:
         def inner():
-            for x, y in zip(self, second):
-                yield result_selector(x, y)
+            yield from zip(self, *iters)
+        return Enumerable(inner)
+
+    def zip2(self, *iters_and_result_selector: Any) -> Enumerable[Any]:
+        iters = iters_and_result_selector[:-1]
+        result_selector = iters_and_result_selector[-1]
+        def inner():
+            for tup in zip(self, *iters):
+                yield result_selector(*tup)
         return Enumerable(inner)
 
     def elements_in(self, *args) -> Enumerable[TSource_co]:
